"""
Variable manager, containing methods to set, get and delete variables
"""
import dataclasses
import enum
import warnings

__all__ = (
    "VariableType", "VariableTypeType",
    "VariableTypesTuple", "VariableInfo",
    "VariableManager", "NORMAL",
    "REQUIRED", "USED_BY_PAYLOAD",
    "VALID_STRING", "VALID_ALPHA",
    "VALID_ALPHANUM", "VALID_ASCII",
    "VALID_PATH_COMPONENT", "VALID_LOCAL_PATH",
    "VALID_HOST_PATH", "VALID_PORT",
    "VALID_IP", "VALID_IPV4",
    "VALID_IPV6"
)

# Type that matches all possible variable types
VariableType = str | int | bool | list[str]
# Type of type that matches all possible variable types
VariableTypeType = type[str | int | bool | list[str]]
# Tuple that matches all possible variable types
VariableTypesTuple = (str, int, bool, list)


class VariableFlags(enum.Flag):
    """
    Flag variables
    """
    NORMAL = 0
    REQUIRED = enum.auto()
    USED_BY_PAYLOAD = enum.auto()
    VALID_STRING = enum.auto()
    VALID_STR_ALPHA = enum.auto()
    VALID_STR_ALPHANUM = enum.auto()
    VALID_STR_ASCII = enum.auto()
    VALID_STR_PATH_COMPONENT = enum.auto()
    VALID_STR_LOCAL_PATH = enum.auto()
    VALID_STR_HOST_PATH = enum.auto()
    VALID_PORT = enum.auto()
    VALID_IP = enum.auto()
    VALID_IPV4 = enum.auto()
    VALID_IPV6 = enum.auto()
    VALID_ALPHA = VALID_STRING | VALID_STR_ALPHA
    VALID_ALPHANUM = VALID_STRING | VALID_STR_ALPHANUM
    VALID_ASCII = VALID_STRING | VALID_STR_ASCII
    VALID_PATH_COMPONENT = VALID_STRING | VALID_STR_PATH_COMPONENT
    VALID_LOCAL_PATH = VALID_STRING | VALID_STR_LOCAL_PATH
    VALID_HOST_PATH = VALID_STRING | VALID_STR_HOST_PATH


# Variable does not need any special treatment
NORMAL = VariableFlags.NORMAL
# Variable is required
REQUIRED = VariableFlags.REQUIRED
# Variable is used by a payload and should not be removed by shell commands
USED_BY_PAYLOAD = VariableFlags.USED_BY_PAYLOAD
# Value must be non-empty, and not be made of entirely whitespace and not contain null bytes
VALID_STRING = VariableFlags.VALID_STRING
# Value must be a valid [a-zA-Z]
VALID_ALPHA = VariableFlags.VALID_ALPHA
# Value must be a valid [a-zA-Z0-9]
VALID_ALPHANUM = VariableFlags.VALID_ALPHANUM
# Value must be a valid ASCII string
VALID_ASCII = VariableFlags.VALID_ASCII
# Value must be a valid path component
VALID_PATH_COMPONENT = VariableFlags.VALID_PATH_COMPONENT
# Value must be an existing, readable path in the local machine
VALID_LOCAL_PATH = VariableFlags.VALID_LOCAL_PATH
# Value must be an existing, readable path in the host machine
VALID_HOST_PATH = VariableFlags.VALID_HOST_PATH
# Value must be a valid port number
VALID_PORT = VariableFlags.VALID_PORT
# Value must be a valid IP (Whatever  IPv4 or IPv6)
VALID_IP = VariableFlags.VALID_IP
# Value must be a valid IPv4
VALID_IPV4 = VariableFlags.VALID_IPV4
# Value must be a valid IPv6
VALID_IPV6 = VariableFlags.VALID_IPV6


@dataclasses.dataclass
class VariableInfo:  # pylint: disable=too-few-public-methods
    """Variable's info
    """
    description: str = ""
    length: int = -1
    var_type: VariableTypeType = str
    flags: VariableFlags = NORMAL
    default_value: VariableType | None = None


class VariableManager:
    """Variable manager, containing methods to set, get and delete variables
    """
    _variables: dict[str, tuple[VariableType, VariableInfo]]
    _VariableValueType = tuple[VariableType, VariableInfo]

    def __init__(self,
                 pre: dict[str, _VariableValueType] | None = None):
        if pre is None:
            self._variables = {}
        else:
            self._variables = {**pre}

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(self._variables)})"

    def has_variable(self, name: str) -> bool:
        """Check if variable is defined

        Args:
            name: Name of variable to check

        Returns:
            True if variable is defined False
        """
        return name in self._variables

    def set_variable(self, name: str, value: VariableType,
                     info_length: int = -1, info_description: str = "",
                     flags: VariableFlags = VariableFlags.NORMAL) -> None:
        """Set value of variable.

        Args:
            name: Name of variable to set
            value: Value of variable
            info_length: Length of variable if -1 no length will be set
            info_description: Description of variable
            flags: Flags for the variable
        """
        if isinstance(value, VariableTypesTuple):
            if isinstance(value, list):
                for i, element in enumerate(value):
                    if not isinstance(element, VariableTypesTuple):
                        raise ValueError(
                            f"Unsupported type of value[{i}]")
                value = list(value)
        else:
            raise ValueError(f"Unsupported type '{type(value)}'")

        if type(value) in (int, bool) and info_length != -1:
            warnings.warn("Variables of integer types cannot have length")
            info_length = -1

        if name in self._variables and not flags & USED_BY_PAYLOAD:
            info = self._variables[name][1]
            if not isinstance(value, info.var_type):
                raise TypeError(
                    f"'{value}' is not of type defined for variable")
            if info.length != -1 and isinstance(value, (str, list)):
                if len(value) > info.length:
                    warnings.warn(
                        f"Value does not fit the size ({info.length})")
                    info_length = -1
            if info_length != -1:
                info.length = info_length
            if len(info_description) == 0:
                info.description = info_description
        else:
            info = VariableInfo()
            info.length = info_length
            info.description = info_description
            info.var_type = type(value)
            info.flags = flags
            if flags & USED_BY_PAYLOAD:
                info.default_value = value
        self._variables[name] = (value, info)

    def del_variable(self, name: str) -> None:
        """Delete a variable from the variable map

        Args:
            name: Name of the variable to delete
        """
        if name in self._variables:
            del self._variables[name]

    def get_variable(self, name: str) -> VariableType:
        """Get the variable with the given name

        Args:
            name: The name of the variable to get

        Returns:
            The variable with the given name
        """
        if name in self._variables:
            return self._variables[name][0]
        raise ValueError(f"No variable named '{name}'")

    def get_variable_str(self, name: str) -> str:
        """Get the variable with the given name, interpreting it as a string

        Args:
            name: The name of the variable to get.

        Raises:
            TypeError: The variable is not meant to be interpreted as a string

        Returns:
            The variable with the given name
        """
        if self.get_variable_info(name).var_type != str:
            raise TypeError(
                f"Variable '{name}' is not to be interpreted as a string")
        return str(self.get_variable(name))

    def get_variable_list(self, name: str) -> list[str]:
        """Get the variable with the given name, interpreting it as a string

        Args:
            name: The name of the variable to get.

        Raises:
            TypeError: The variable is not meant to be interpreted as a list

        Returns:
            The variable with the given name
        """
        if self.get_variable_info(name).var_type != list:
            raise TypeError(
                f"Variable '{name}' is not to be interpreted as a list")
        value = self.get_variable(name)
        assert isinstance(value, list)
        return value

    def get_variable_info(self, name: str) -> VariableInfo:
        """Get information about a variable

        Args:
            name: Name of the variable to get information about

        Raises:
            KeyError: Name is not a recognized variable

        Returns:
            Variable information
        """
        if name in self._variables:
            return self._variables[name][1]
        raise KeyError(f"No variable named '{name}'")

    def get_variables_name(self) -> list[str]:
        """Get list of variables names.

        Returns:
            list of variable names
        """
        return list(self._variables.keys())

    def get_variables_value(self) -> list[VariableType]:
        """Get list of variables values.

        Returns:
            list of variable values
        """
        return list(v[0] for v in self._variables.values())

    def get_variables_info(self) -> list[VariableInfo]:
        """Get list of all variable's info

        Returns:
            list of all variable info
        """
        return list(v[1] for v in self._variables.values())



"""
Change registry on the device. Add registries like this:

set HKLM\\Whatever\\Whatever2 key:type:value

To set the default key, value for a registry. Use the defacto standard of @ for key:

set HKLM\\Whatever\\Whatever @:type:value

Valid types are:

    REG_BINARY, REG_NONE: Enter as hex value (e.g 00 01 02)
    REG_DWORD, REG_DWORD_BIG_ENDIAN, REG_QWORD: Enter as integer (e.g 127)
    REG_EXPAND_SZ, REG_LINK, REG_SZ: Enter as string (e.g my value)

The payload will automatically search for variables that start with HKLM or HKEY_LOCAL_MACHINE and process them.
"""
from snr.core.payload.payload import Context, Payload
from snr.core.util import common_utils
from snr.cli import variables


class RegistryPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Microsoft Windows",)
    INPUTS = (
        Payload.supports_encrypted_access()
    )
    VALID_REG_TYPEs: dict[str, type[bytes | int | str]] = {
        "REG_BINARY": bytes,
        "REG_NONE": bytes,
        "REG_DWORD": int,
        "REG_DWORD_BIG_ENDIAN": int,
        "REG_QWORD": int,
        "REG_EXPAND_SZ": str,
        "REG_LINK": str,
        "REG_SZ": str,
    }

    def generate(self, ctx: Context) -> int:
        registries: dict[str, bytes | int | str] = {}
        our_vars = {}
        for name in variables.global_vars.get_variables_name():
            if name.startswith(("HKLM", "HKEY_LOCAL_MACHINE")):
                # Both are common mistakes we can deal with
                name = name.replace("\\\\", "\\").replace("/", "\\")
                # While they are technically correct in syntax, it will break our payload
                name = name.replace("HKLM:", "HKLM\\").replace("HKEY_LOCAL_MACHINE:", "HKEY_LOCAL_MACHINE\\")
                # Now that we know this variable is intended for us, we can try to validate it
                try:
                    key, type_, value = variables.global_vars.get_variable_str(
                        name).split(":", maxsplit=2)
                except ValueError:
                    common_utils.print_error(
                        f"Invalid syntax for value of {name}, it must be in format key:type:value")
                    return 1
                except TypeError:
                    common_utils.print_error(
                        f"Invalid type for variable {name}. The variable must be a string")
                    return 1
                if len(key) == 0:
                    common_utils.print_error(
                        f"Invalid key for {name}. For default value use @ instead of an empty key")
                    return 1
                type_ = type_.upper()
                if type_ not in self.VALID_REG_TYPEs:
                    common_utils.print_error(
                        f"Invalid type for {name}. Use `info` to see more about valid types")
                    return 1
                # Interpret and validate the value
                new_value: bytes | int | str
                match type_:
                    case "REG_BINARY" | "REG_NONE":
                        # Value is a sequence of hex bytes separated by spaces
                        new_value = b""
                        for byte in value.split(" "):
                            try:
                                new_value += bytes(int(byte, base=16))
                            except ValueError:
                                common_utils.print_error(
                                    f"Invalid value for {name}. '{byte}' is not a valid hexadecimal number")
                                return 1

                    case "REG_DWORD" | "REG_DWORD_BIG_ENDIAN" | "REG_QWORD":
                        # Value is a number
                        try:
                            new_value = int(value)
                        except ValueError:
                            common_utils.print_error(
                                f"Invalid value for {name}. '{value}' is not a valid hexadecimal number")
                            return 1
                        # We don't concern ourselves with the integer size, that's not our issue
                    case "REG_EXPAND_SZ" | "REG_LINK" | "REG_SZ":
                        new_value = value
            registries[name + f":{key}:{type_}"] = new_value
        our_vars["REGISTRIES"] = registries
        self.format_payload_and_write(ctx, our_vars)
        self.add_autorun(ctx)
        return 0


payload = RegistryPayload()



"""
Access and copy files by pattern
"""

from snr.core.payload.payload import Context, Payload, REQUIRED


class FilesPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Microsoft Windows", "GNU/Linux")
    INPUTS = (
        ("PATTERNS", [], -1, "Glob pattern for files to copy", REQUIRED),
        Payload.supports_encrypted_access()
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        self.format_payload_and_write(ctx, variables)
        self.add_autorun(ctx)
        return 0


payload = FilesPayload()



"""
Run an executable on boot, the executable will be copied to the host filesystem.
If you want to run an executable that is already available on the host. Use run_command.
It does not copy the executable onto the host filesystem.
Do note that this payload does preserve ordering
"""
import os.path
import shutil

from snr.core.payload.payload import Context, Payload, REQUIRED, VALID_STRING
from snr.core.util import common_utils


class RunCommandPayload(Payload):
    AUTHORS = ("GlobularOne",)
    TARGET_OS_LIST = ("Any",)
    INPUTS = (
        ("EXECUTABLES", [], -1, "Executables to copy and run", REQUIRED | VALID_STRING),
    )

    def generate(self, ctx: Context) -> int:
        variables = self.get_self_variables()
        assert isinstance(variables["EXECUTABLES"], list)
        if len(variables["EXECUTABLES"]) == 0:
            common_utils.print_error("Nothing selected to run")
            return 1
        for executable in variables["EXECUTABLES"]:
            basename = os.path.basename(executable)
            target = os.path.join("root", basename)
            if not os.path.isabs(executable):
                # Does it exist on the current working directory?
                if os.path.exists(os.path.abspath(executable)):
                    # Then it's fine
                    executable = os.path.abspath(executable)
                else:
                    # Does it exist on PATH?
                    tmp = shutil.which(executable)
                    if tmp is None:
                        # We can't find it
                        common_utils.print_error(
                            f"Could not find '{executable}'! Check that you have typed correctly")
                        return 1
                    else:
                        executable = tmp

            common_utils.print_debug(f"Copying '{executable}' to '{target}'")
            try:
                shutil.copyfile(executable, ctx.join(target))
            except (OSError) as exc:
                common_utils.print_error(
                    f"Installing executable to rootfs failed: {exc}")
                return 1
            common_utils.print_debug(
                f"Adding autorun service for '{executable}'")
            self.add_autorun(ctx, os.path.join("/", target))
        return 0


payload = RunCommandPayload()



"""
Shell variables
"""
from snr.core.core import variable_manager
from snr.core.core.variable_manager import (NORMAL, REQUIRED, USED_BY_PAYLOAD,
                                            VALID_ALPHA, VALID_ALPHANUM,
                                            VALID_ASCII, VALID_HOST_PATH,
                                            VALID_PORT, VALID_IP,
                                            VALID_IPV4, VALID_IPV6,
                                            VALID_LOCAL_PATH,
                                            VALID_PATH_COMPONENT, VALID_STRING)

__all__ = (
    "global_vars", "NORMAL",
    "REQUIRED", "USED_BY_PAYLOAD",
    "VALID_STRING", "VALID_ALPHA",
    "VALID_ALPHANUM", "VALID_ASCII",
    "VALID_PATH_COMPONENT", "VALID_LOCAL_PATH",
    "VALID_HOST_PATH", "VALID_PORT",
    "VALID_IP", "VALID_IPV4",
    "VALID_IPV6"
)

global_vars = variable_manager.VariableManager()



"""
Edit NT registry values
"""
import contextlib
import struct
from typing import Any, Generator

import hivex
import hivex.hive_types
from hivex.hive_types import (REG_BINARY, REG_DWORD, REG_DWORD_BIG_ENDIAN,
                              REG_EXPAND_SZ, REG_LINK, REG_MULTI_SZ, REG_NONE,
                              REG_QWORD, REG_SZ)

from snr.core.core import path_wrapper

__all__ = (
    "REG_BINARY", "REG_DWORD", "REG_DWORD_BIG_ENDIAN",
    "REG_EXPAND_SZ", "REG_LINK", "REG_MULTI_SZ",
    "REG_NONE", "REG_QWORD", "REG_SZ",
    "RegistryNodeValue", "RegistryNode", "NtRegistry",
)

# Mapping of registry types to python types
REG_TYPE_MAPPING: dict[int, type[bytes | int | str | list]] = {
    REG_BINARY: bytes,
    REG_DWORD: int,
    REG_DWORD_BIG_ENDIAN: int,
    REG_EXPAND_SZ: str,
    REG_LINK: str,
    REG_MULTI_SZ: list,
    REG_NONE: bytes,
    REG_QWORD: int,
    REG_SZ: str,
}


class RegistryNodeValue:
    """
    A value of a node in a registry
    """
    _hive: hivex.Hivex
    _node_handle: int
    _key: str
    _value: bytes | int | str | list[str]
    _type: int

    def __init__(self, hive: hivex.Hivex, node_handle: int, key: str, type_value: tuple[int, bytes | int | str | list[str] | None] = (REG_NONE, None)):
        self._hive = hive
        self._node_handle = node_handle
        self._key = key
        if type_value[1] is None:
            self._type, value = self._hive.value_value(
                self._hive.node_get_value(self._node_handle, self.key))
        else:
            self._type, value = type_value
        # Try to interpret the value
        match self._type:
            case hivex.hive_types.REG_BINARY:
                self._value = value
            case hivex.hive_types.REG_DWORD:
                self._value = struct.unpack("<I", value)[0]
            case hivex.hive_types.REG_DWORD_BIG_ENDIAN:
                self._value = struct.unpack(">I", value)[0]
            case hivex.hive_types.REG_EXPAND_SZ:
                self._value = value.decode()
            case hivex.hive_types.REG_LINK:
                self._value = value.decode()
            case hivex.hive_types.REG_MULTI_SZ:
                self._value = []
                values = value.split(b"\x00")
                if len(values) > 2:
                    # By definition the sequence comes with it's own null terminator
                    for i in range(len(values) - 2):
                        self._value.append(values[i].decode())
            case hivex.hive_types.REG_NONE:
                # We can't guess the type, so we don't process it
                self._value = value
            case hivex.hive_types.REG_QWORD:
                self._value = struct.unpack("<Q", value)[0]
            case hivex.hive_types.REG_SZ:
                self._value = value.decode()
            case _:
                # We don't know it
                self._value = value
        if type_value[1] is None:
            self.sync()

    def sync(self):
        """
        Synchronize our information with the registry, note that this doesn't change it on the disk.
        You generally don't need to worry about this and just change the value.
        """
        packed_value: bytes
        match self._type:
            case hivex.hive_types.REG_BINARY:
                assert isinstance(self._value, bytes)
                packed_value = self._value
            case hivex.hive_types.REG_DWORD:
                assert isinstance(self._value, int)
                packed_value = struct.pack("<I", self._value)
            case hivex.hive_types.REG_DWORD_BIG_ENDIAN:
                assert isinstance(self._value, int)
                packed_value = struct.pack(">I", self._value)
            case hivex.hive_types.REG_EXPAND_SZ:
                assert isinstance(self._value, str)
                packed_value = self._value.encode() + b"\x00"
            case hivex.hive_types.REG_LINK:
                assert isinstance(self._value, str)
                packed_value = self._value.encode() + b"\x00"
            case hivex.hive_types.REG_MULTI_SZ:
                assert isinstance(self._value, list)
                packed_value = b""
                for value in self._value:
                    assert isinstance(value, str)
                    packed_value += value.encode() + b"\x00"
                packed_value += b"\x00"

            case hivex.hive_types.REG_NONE:
                assert isinstance(self._value, bytes)
                # We didn't process it when loading, we won't process it while dumping neither
                packed_value = self._value
            case hivex.hive_types.REG_QWORD:
                assert isinstance(self._value, int)
                packed_value = struct.pack("<Q", self._value)
            case hivex.hive_types.REG_SZ:
                assert isinstance(self._value, str)
                packed_value = self._value.encode() + b"\x00"
            case _:
                assert isinstance(self._value, bytes)
                packed_value = self._value
        self._hive.node_set_value(self._node_handle, {
            "key": self._key,
            "t": self._type,
            "value": packed_value
        })

    @property
    def key(self) -> str:
        """Key of the value
        """
        return self._key

    @property
    def value(self) -> bytes | int | str | list[str]:
        """Registry value
        """
        return self._value

    @value.setter
    def value(self, new_value:  bytes | int | str | list[str]):
        if self._type not in REG_TYPE_MAPPING:
            if not isinstance(new_value, bytes):
                raise TypeError(
                    "Unknown type registry value must be of type bytes")
        elif not isinstance(new_value, REG_TYPE_MAPPING[self._type]):
            raise TypeError(
                f"Value must of type: {REG_TYPE_MAPPING[self._type].__name__}")
        self._value = new_value
        self.sync()

    @property
    def type(self) -> int:
        """Type of the registry value
        """
        return self._type

    def value_as_bytes(self) -> bytes:
        """Return value interpreting it as bytes

        Raises:
            TypeError: Value is not meant to be interpreted as bytes
        """
        if REG_TYPE_MAPPING[self._type] != bytes:
            raise TypeError("Value cannot be interpreted as bytes")
        assert isinstance(self._value, bytes)
        return self._value

    def value_as_string(self) -> str:
        """Return value interpreting it as a string

        Raises:
            TypeError: Value is not meant to be interpreted as a string
        """
        if REG_TYPE_MAPPING[self._type] != str:
            raise TypeError("Value cannot be interpreted as a string")
        assert isinstance(self._value, str)
        return self._value

    def value_as_int(self) -> int:
        """Return value interpreting it as an integer

        Raises:
            TypeError: Value is not meant to be interpreted as an integer
        """
        if REG_TYPE_MAPPING[self._type] != int:
            raise TypeError("Value cannot be interpreted as an integer")
        assert isinstance(self._value, int)
        return self._value

    def value_as_list(self) -> list[str]:
        """Return value interpreting it as a list of strings

        Raises:
            TypeError: Value is not meant to be interpreted as a list of strings
        """
        if REG_TYPE_MAPPING[self._type] != list:
            raise TypeError("Value cannot be interpreted as a list of strings")
        assert isinstance(self._value, list)
        return self._value


class RegistryNode:
    """A registry node

    Attributes:
        handle: Handle of the node
        name: Name of the node
        Timestamp: Last modification timestamp
        parent: Parent node of this node or none if it's the root node
        data: Key and values assigned to this node
    """
    _hive: hivex.Hivex
    handle: int
    name: str
    timestamp: int
    parent: 'RegistryNode' | None
    data: dict[str, RegistryNodeValue]

    def __init__(self, hive: hivex.Hivex, handle: int):
        self._hive = hive
        self.handle = handle
        # Query info
        self.name = self._hive.node_name(self.handle)
        self.timestamp = self._hive.node_timestamp(self.handle)
        try:
            self.parent = self._hive.node_parent(self.handle)
        except RuntimeError:
            # No parent, we are the root
            self.parent = None
        for data_handle in self._hive.node_values(self.handle):
            key = self._hive.value_key(data_handle)
            self.data[key] = RegistryNodeValue(self._hive, self.handle, key)
        self._hive.node_set_value()

    def get_children_count(self) -> int:
        """Get count of children this node has

        Returns:
            Count of children of the current node
        """
        return self._hive.node_nr_children(self.handle)

    def get_children(self) -> Generator['RegistryNode', None, None]:
        """Get children of this node

        Yields:
            Child of this node
        """
        for i in self._hive.node_children(self.handle):
            yield RegistryNode(self._hive, i)

    def find_child(self, key: str) -> 'RegistryNode' | None:
        """Find a child by it's name

        Args:
            key: Name of the child

        Returns:
            The node if found, none otherwise
        """
        return self._hive.node_get_child(self.handle, key)

    def new_child(self, name: str) -> 'RegistryNode':
        """Create a new child node and return it
        """
        return RegistryNode(self._hive, self._hive.node_add_child(self.handle, name))

    def del_child(self, node: 'RegistryNode'):
        """

        Args:
            node: Node to delete
        """
        self._hive.node_delete_child(node.handle)

    def new_data(self, key: str, type_: int, value: bytes | int | str | list[str]) -> RegistryNodeValue:
        """Create a new key, data pair and return it

        Args:
            key: Key for the value
            type_: Type fo the value
            value: Note that value must be compatible with the type

        Returns:
            The registry node value
        """
        return RegistryNodeValue(self._hive, self.handle, key, (type_, value))


class NtRegistry(contextlib.AbstractContextManager['NtRegistry']):
    """
    Manager registry hives
    """
    _hives: dict[str, hivex.Hivex]
    _path_wrapper: path_wrapper.PathWrapperBase

    def __init__(self, wrapper: path_wrapper.PathWrapperBase):
        self._hives = {}
        self._path_wrapper = wrapper

    def __enter__(self) -> 'NtRegistry':
        return self

    def __exit__(self, *_: Any) -> None:
        # To exit, commit all hives and close them
        for hive_name, hive in self._hives.values():
            hive.commit(f"Windows/System32/config/{hive_name}")
            del self._hives[hive_name]

    def find_node(self, path: str, root_node: RegistryNode | None = None) -> RegistryNode | None:
        """Find a node

        There are a few possible valid syntax for path here:
            (HKLM|HKEY_LOCAL_MACHINE)\\whatever
            (HKLM|HKEY_LOCAL_MACHINE):\\whatever
            \\whatever
        The first two do not need a root_node, we are going to find it ourselves (they are absolute)
        The third syntax is a relative search and root_node cannot be None.

        Args:
            path: Path to the node to find
            root_node: Root node to start search from, only used for relative searches. Defaults to None.

        Raises:
            ValueError: If we cannot find the hive or requesting a relative search without a root_node
            FileNotFoundError: We do know the hive but can't find it

        Returns:
            The node found or None if not found
        """
        if path.startswith(("HKLM", "HKEY_LOCAL_MACHINE")):
            # Absolute path
            path = path.removeprefix("HKLM").removeprefix("HKEY_LOCAL_MACHINE")
            path = path.removeprefix("\\").removeprefix(":")
            hive_name = path.split("\\")[0].upper()
            if hive_name not in ("SOFTWARE", "SECURITY", "SYSTEM", "SAM"):
                # we don't know that
                raise ValueError(f"Unknown hive '{hive_name}'")
            # First check if already have opened it
            if hive_name in self._hives:
                hive = self._hives[hive_name]
            else:
                # We have not opened this hive yet, do so
                hive_path = f"Windows/System32/config/{hive_name}"
                if not self._path_wrapper.exists(hive_path):
                    raise FileNotFoundError(
                        f"File '{self._path_wrapper.join(hive_path)}' not found")
                hive = hivex.Hivex(hive_path, write=True)
                self._hives[hive_name] = hive
            root_node = RegistryNode(hive, hive.root())
            # Make the path relative
            path = path.removeprefix(path.split("\\")[0])
        elif path.startswith("\\"):
            if root_node is None:
                raise ValueError(
                    "Cannot do a relative search without a root node")
        else:
            raise ValueError("Path is of unknown format")
        # Make the path suitable for search
        path = path.removeprefix("\\")
        parent_node = root_node
        for node_name in path.split("\\"):
            new_node = parent_node.find_child(node_name)
            if new_node is None:
                return None
            parent_node = new_node
        return parent_node
